/*
 * CopyrightC NEC Corporation, NEC CustomTechnica, Ltd. 2003
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
//#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>
#include "common.h"
#include "libapm.h"
#include "libapm_rw.h"
#include "myerror.h"
#include "debug.h"

extern "C" {
#include "apm.h"
}

#define MAX_EVENTS			 8		/* Maximum events we accept from BIOS driver */
#define CHECK_POWER_TIME	-1		/* seconds between /proc/apm checks */
#define NEC_FUNC_ON			 1
#define NEC_FUNC_OFF		 0
#define APM_START			 1
#define APM_STOP			 0
#define SUCCESS				 0		// ؿֵ
#define NOT_SUCCESS			 1		// ؿ۾ֵ

#define	LOGFILE_NAME		"mcapm.log"

typedef struct {
	int fd;
	int apm_fd;
} APM_FORMAT;

CErrorCode	apm_error;

int reallocate(unsigned char **p1, unsigned char *p2, int size)
{
	unsigned char *dest = (unsigned char *)malloc(sizeof(FIXED_FORMAT) + size);

	if (dest == NULL) return 0;

	memcpy(dest, *p1, sizeof(FIXED_FORMAT));
	memcpy(dest+sizeof(FIXED_FORMAT), p2, size);
	free(*p1);
	*p1 = dest;
	return (sizeof(FIXED_FORMAT) + size);
}


// ׵ѥåԽ
void Edit_ReqFixData(FIXED_FORMAT *w_fbuf, unsigned char toid, unsigned char command, unsigned char status, unsigned short datatype, unsigned short datalength)
{
	memset(w_fbuf, 0, FIXED_AREA_SIZE);

	w_fbuf->to.id      = toid;
	w_fbuf->from.id    = MID_APM;
	w_fbuf->command    = command;
	w_fbuf->status     = status;
	w_fbuf->datatype   = datatype;
	w_fbuf->datalength = datalength;

    return;
}


// ManagerؤΥ顼ѥѥåԽ
void CmdErrorData_Edit(FIXED_FORMAT *buf, unsigned char status)
{
	memset(buf, 0, FIXED_AREA_SIZE);

	buf->to.id    = MID_MGR;		// to.id
	buf->from.id  = MID_APM;			// from.id
	buf->command  = CMD_ERROR;		// command
	buf->status   = status;			// status
	buf->datatype = 0x0000;			// datatype

    return;
}


// ѥѥåȸԽ
void Edit_RspFixData(FIXED_FORMAT *r_fbuf, unsigned char status, unsigned short datatype, unsigned short datalength, FIXED_FORMAT *w_fbuf, int *w_len)
{
	ADDR	tmp;

	memset(w_fbuf, 0, FIXED_AREA_SIZE);

	// ѸԽ
	memcpy(&tmp,          &r_fbuf->to,   sizeof(ADDR));
	memcpy(&w_fbuf->to,   &r_fbuf->from, sizeof(ADDR));		// to
	memcpy(&w_fbuf->from, &tmp,          sizeof(ADDR));		// from
	w_fbuf->command    = r_fbuf->command | RESPONSE_FLG;	// command
	w_fbuf->status     = status;							// status
	w_fbuf->datatype   = datatype;							// datatype
	w_fbuf->datalength = datalength;						// datalength

	*w_len = FIXED_AREA_SIZE;

	return;
}


void *check_apm_event(void *arg)
{
	APM_FORMAT		*pfd;
	apm_info		apminfo, *apmi;
	apm_event_t		events[MAX_EVENTS];
	FIXED_FORMAT	w_buf;		 	// ѥХåե
	int				res;

	TraceDisp1("APM(%d) : check_apm_event() start", __LINE__);

	// åɤΥ󥻥륿׻ꡧƱ
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);	// 2002.12.27 t.masuzawa

	pfd = (APM_FORMAT *)arg;

	for (;;) {
		int i;
		int n = apm_get_events(pfd->apm_fd, CHECK_POWER_TIME, events, MAX_EVENTS);

		for (i = 0; i < n; i++) {
			if (apm_read(&apminfo)) {
				continue;
			}
			apmi = &apminfo;
			TraceDisp2("APM(%d) : event = %04x", __LINE__, events[i]);
			if (events[i] == APM_OEM_NEC_1) {
				TraceDisp1("APM(%d) : Power Button Pushed !!", __LINE__);
				apm_error.OutputLog("APM(%d) : Power Button Pushed !!", __LINE__);

				Edit_ReqFixData(&w_buf, MID_MGR, MGR_NOTICE_POWER_ONOFF, STS_SUCCESS, 0, 0);

				PrintFormat((unsigned char *)&w_buf);

				res = apm_Send_Data(pfd->fd, (unsigned char *)&w_buf, sizeof(FIXED_FORMAT));

				if (res != SUCCESS) {
					TraceDisp1("APM(%d) : check_apm_event() error (%04X)", __LINE__, res);
					apm_error.OutputLog("APM(%d) : check_apm_event() error (%04X)", __LINE__, res);
				}
			}
		}
	}

	TraceDisp1("APM(%d) : check_apm_event() end", __LINE__);

	pthread_exit(0);
}


int set_nec_function(int apm_fd, int mode)
{
	int retval;
	int errno = 0;

	TraceDisp1("APM(%d) : set_apm_event() start", __LINE__);

	if (mode == NEC_FUNC_OFF) {
		retval = ioctl(apm_fd, APM_IOC_DISABLE_OEM_NEC_1, NULL);
	} else if (mode == NEC_FUNC_ON) {
		retval = ioctl(apm_fd, APM_IOC_ENABLE_OEM_NEC_1, NULL);
	} else {
		TraceDisp1("APM(%d) : illegal parameter (mode=%d)", __LINE__, mode);
		TraceDisp1("APM(%d) : set_apm_event() end", __LINE__);
		apm_error.SetErrorCode(retval, __FILE__, __LINE__, errno);
		apm_error.OutputLog("APM(%d) : illegal parameter (mode=%d)", __LINE__, mode);
		return NOT_SUCCESS;
	}

	if (retval == -1) {
		TraceDisp1("APM(%d) : cannot set nec_function (mode=%d)", __LINE__, mode);
		TraceDisp1("APM(%d) : set_apm_event() end", __LINE__);
		apm_error.SetErrorCode(retval, __FILE__, __LINE__, errno);
		apm_error.OutputLog("APM(%d) : illegal parameter (mode=%d)", __LINE__, mode);
		return NOT_SUCCESS;
	}

	TraceDisp1("APM(%d) : set_apm_event() end", __LINE__);
	return SUCCESS;
}


int get_resume_timer(int apm_fd, struct apm_time *apm_tm)
{
	int retval, res = SUCCESS;
	int errno = 0;

	TraceDisp1("APM(%d) : get_resume_timer() start", __LINE__);

	/* Read back the current alarm settings */
	retval = ioctl(apm_fd, APM_IOC_GET_RESUME_TIMER, apm_tm);
	if (retval == -1) {
		TraceDisp1("APM(%d) : ioctl(APM_IOC_GET_RESUME_TIMER) failed", __LINE__);
		apm_error.SetErrorCode(retval, __FILE__, __LINE__, errno);
		apm_error.OutputLog("APM(%d) : ioctl(APM_IOC_GET_RESUME_TIMER) failed", __LINE__);
		res = NOT_SUCCESS;
	} else {
//		TraceDisp3("APM(%d) : APM alarm time is %04d-%02d-%02d %02d:%02d:%02d", __LINE__,
//							apm_tm->tm_year, apm_tm->tm_mon, apm_tm->tm_mday,
//							apm_tm->tm_hour, apm_tm->tm_min, apm_tm->tm_sec);
		printf("APM(%d) : APM alarm time is %04d-%02d-%02d %02d:%02d:%02d", __LINE__,
							apm_tm->tm_year, apm_tm->tm_mon, apm_tm->tm_mday,
							apm_tm->tm_hour, apm_tm->tm_min, apm_tm->tm_sec);
		apm_error.OutputLog("APM(%d) : APM alarm time is %04d-%02d-%02d %02d:%02d:%02d", __LINE__,
							apm_tm->tm_year, apm_tm->tm_mon, apm_tm->tm_mday,
							apm_tm->tm_hour, apm_tm->tm_min, apm_tm->tm_sec);
	}

	TraceDisp1("APM(%d) : get_resume_timer() end", __LINE__);

	return res;
}


int set_resume_timer(int apm_fd, struct apm_time *apm_tm)
{
	int retval, res = SUCCESS;
	int errno;
	struct tm kernel_tm;
	time_t kernel_time;

	TraceDisp1("APM(%d) : set_resume_timer() start", __LINE__);

	time(&kernel_time);
	kernel_tm=*localtime(&kernel_time);

	TraceDisp3("APM(%d) : Current local time is %04d-%02d-%02d %02d:%02d:%02d", __LINE__,
					kernel_tm.tm_year + 1900, kernel_tm.tm_mon + 1, kernel_tm.tm_mday,
					kernel_tm.tm_hour, kernel_tm.tm_min, kernel_tm.tm_sec);

	TraceDisp3("APM(%d) : Setting APM alarm time to %04d-%02d-%02d %02d:%02d:%02d", __LINE__,
							apm_tm->tm_year, apm_tm->tm_mon, apm_tm->tm_mday,
							apm_tm->tm_hour, apm_tm->tm_min, apm_tm->tm_sec);

	apm_error.OutputLog("APM(%d) : Current local time is %04d-%02d-%02d %02d:%02d:%02d", __LINE__,
					kernel_tm.tm_year + 1900, kernel_tm.tm_mon + 1, kernel_tm.tm_mday,
					kernel_tm.tm_hour, kernel_tm.tm_min, kernel_tm.tm_sec);

	apm_error.OutputLog("APM(%d) : Setting APM alarm time to %04d-%02d-%02d %02d:%02d:%02d", __LINE__,
							apm_tm->tm_year, apm_tm->tm_mon, apm_tm->tm_mday,
							apm_tm->tm_hour, apm_tm->tm_min, apm_tm->tm_sec);

	retval = ioctl(apm_fd, APM_IOC_SET_RESUME_TIMER, apm_tm);
	if (retval == -1) {
		TraceDisp1("APM(%d) : ioctl(APM_IOC_SET_RESUME_TIMER) failed", __LINE__);
		apm_error.SetErrorCode(retval, __FILE__, __LINE__, errno);
		apm_error.OutputLog("APM(%d) : ioctl(APM_IOC_SET_RESUME_TIMER) failed", __LINE__);
		res = NOT_SUCCESS;
	}

	TraceDisp1("APM(%d) : set_resume_timer() end", __LINE__);

	return res;
}


void *apm_main(void *arg)
{
	int				start_flg;		// ⥸塼뵯ư
	int				res;			// ؿ
	char			rsp_sts;		// ѥơ
	FIXED_FORMAT	r_fbuf;			// ɤ߹ѥХåե()
	unsigned char	*r_vbuf = NULL;	// ɤ߹ѥХåե()
	unsigned char	*w_buf = NULL; 	// 񤭹ѥХåե
	unsigned char	*w_vbuf = NULL; 	// 񤭹ѥХåե
	int				w_len; 			// 񤭹ѥѥåĹ
	int				loop_sts;		// ᥤ롼׾
	int				errno = 0;
	APM_FORMAT		fd;
	pthread_t 		check_apm_event_thread;	// Ÿܥƻ륹åID
	char			logfile[MAX_PATH];

	TraceDisp1("APM(%d) : apm_main() start", __LINE__);

	char *tmp = getenv("MC_LOGFILE_PATH");
	if (tmp == NULL) {
		sprintf(logfile, "%s/%s", MC_LOGFILE_PATH, LOGFILE_NAME);
	} else {
		sprintf(logfile, "%s/%s", tmp, LOGFILE_NAME);
	}

	TraceDisp3("APM(%d) : logfile = %s", __LINE__, logfile);

	apm_error.Initialize(logfile, MAX_LOG_LINE, MAX_LOG_COLUMN, MID_APM);

	// ̿ʳΩ
	fd.fd = (int)arg;

	// APM Check
	if (getuid() != 0) {
		TraceDisp("APM(%d) : must be run as root.", __LINE__);
		apm_error.OutputLog("APM(%d) : must be run as root.", __LINE__);
		pthread_exit((void *)1);
	}

	if (apm_exists()) {
		TraceDisp("APM(%d) : No support APM.", __LINE__);
		apm_error.OutputLog("APM(%d) : No support APM.", __LINE__);
		pthread_exit((void *)1);
	}

	if ((fd.apm_fd = apm_open()) < 0) {
		TraceDisp("APM(%d) : apm_open() failed (%d).", __LINE__, fd.apm_fd);
		apm_error.OutputLog("APM(%d) : apm_open faild (%d).", __LINE__, fd.apm_fd);
		pthread_exit((void *)1);
	}

	if (set_nec_function(fd.apm_fd, NEC_FUNC_ON)) {
		TraceDisp("APM(%d) : No support APM.", __LINE__);
		apm_error.OutputLog("APM(%d) : No support APM.", __LINE__);
		pthread_exit((void *)1);
	}

	pthread_create(&check_apm_event_thread, NULL, check_apm_event, (void *)&fd);

	loop_sts = 1;
	start_flg = APM_STOP;

	// ᥤ롼
	while(loop_sts)
	{
		TraceDisp2("APM(%d) : loop start", __LINE__);

		rsp_sts = STS_SUCCESS;

		// ManagerΥǡ
		res = apm_Select_Data(fd.fd, NULL);
		if (res < 0) {
			TraceDisp1("APM(%d) : apm_main() error (%04X)", __LINE__, res);
			TraceDisp1("APM(%d) : apm_main() end", __LINE__);
			apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
			apm_error.OutputLog("APM(%d) : apm_main() error (%04X)", __LINE__, res);
			CmdErrorData_Edit(&r_fbuf, res);
			apm_Send_Data(fd.fd, (unsigned char *)&r_fbuf, FIXED_AREA_SIZE);
			pthread_exit((void *)1);
		}

		// ɤ߹ǡ¤ΤسǼ
		res = apm_Recv_fData(fd.fd, &r_fbuf);
		if (res != SUCCESS) {
			TraceDisp1("APM(%d) : apm_main() error (%04X)", __LINE__, res);
			TraceDisp1("APM(%d) : apm_main() end", __LINE__);
			apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
			apm_error.OutputLog("APM(%d) : apm_main() error (%04X)", __LINE__, res);
			CmdErrorData_Edit(&r_fbuf, STS_READ_ERROR);
			apm_Send_Data(fd.fd, (unsigned char *)&r_fbuf, FIXED_AREA_SIZE);
			pthread_exit((void *)1);
		}

		if (r_fbuf.datalength > 0) {
			// Ĺ顢Υǡΰݤ
			r_vbuf = (unsigned char *)malloc(r_fbuf.datalength);
			if (r_vbuf == NULL) {
				TraceDisp1("APM(%d) : apm_main() malloc error (%04X)", __LINE__, STS_MEMORY_ERROR);
				TraceDisp1("APM(%d) : apm_main() end", __LINE__);
				apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
				apm_error.OutputLog("APM(%d) : apm_main() malloc error (%04X)", __LINE__, STS_MEMORY_ERROR);
				CmdErrorData_Edit(&r_fbuf, STS_MEMORY_ERROR);
				apm_Send_Data(fd.fd, (unsigned char *)&r_fbuf, FIXED_AREA_SIZE);
				pthread_exit((void *)1);
			}

			// ɤ߹ǡΰسǼ
			res = apm_Recv_vData(fd.fd, r_fbuf.datalength, r_vbuf);
			if (res != SUCCESS) {
				TraceDisp1("APM(%d) : apm_main() error (%04X)", __LINE__, res);
				TraceDisp1("APM(%d) : apm_main() end", __LINE__);
				apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
				apm_error.OutputLog("APM(%d) : apm_main() error (%04X)", __LINE__, STS_READ_ERROR);
				CmdErrorData_Edit(&r_fbuf, STS_READ_ERROR);
				apm_Send_Data(fd.fd, (unsigned char *)&r_fbuf, FIXED_AREA_SIZE);
				pthread_exit((void *)1);
			}
		}

		if (r_fbuf.command & RESPONSE_FLG) {
			if (r_vbuf) { free(r_vbuf); r_vbuf = NULL; }
			continue;
		}

		// ΰݡ
		w_buf = (unsigned char *)malloc(FIXED_AREA_SIZE);
		if (w_buf == NULL) {
			TraceDisp1("APM(%d) : apm_main() malloc error (%04X)", __LINE__, STS_MEMORY_ERROR);
			TraceDisp1("APM(%d) : apm_main() end", __LINE__);
			apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
			apm_error.OutputLog("APM(%d) : apm_main() malloc error (%04X)", __LINE__, STS_MEMORY_ERROR);
			CmdErrorData_Edit(&r_fbuf, STS_MEMORY_ERROR);
			apm_Send_Data(fd.fd, (unsigned char *)&r_fbuf, FIXED_AREA_SIZE);
			pthread_exit((void *)1);
		}

		memset(w_buf, (char)NULL, FIXED_AREA_SIZE);

		switch(r_fbuf.command) {
/********** From Manager *********/
			case	CMD_POWERON		:	// ŸON
					TraceDisp2("APM(%d) : CMD_POWERON", __LINE__);
					Edit_RspFixData(&r_fbuf, STS_SUCCESS, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
					break;

			case	CMD_POWEROFF	:	// ŸOFF
					TraceDisp2("APM(%d) : CMD_POWEROFF", __LINE__);
					Edit_RspFixData(&r_fbuf, STS_SUCCESS, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
					break;

			case	CMD_START		:	// ⥸塼볫׵
			case	CMD_START2		:	// ⥸塼볫׵ᣲ
					TraceDisp2("APM(%d) : CMD_START", __LINE__);
					start_flg = APM_START;
					Edit_RspFixData(&r_fbuf, STS_SUCCESS, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
					break;

			case	CMD_SHUT		:	// ⥸塼뽪λ׵
					TraceDisp2("APM(%d) : CMD_START", __LINE__);
					start_flg = APM_STOP;
					res = pthread_cancel(check_apm_event_thread);
					if( res != SUCCESS ) {
						TraceDisp1("APM(%d) : apm_main() pthread_cancel error (%04X)", __LINE__, res);
						apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
						apm_error.OutputLog("APM(%d) : apm_main() pthread_cancel error (%04X)", __LINE__, res);
					} else {
						pthread_join(check_apm_event_thread, NULL);	// 2002.12.27 t.masuzawa
					}
					Edit_RspFixData(&r_fbuf, STS_SUCCESS, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
					// 롼׽λ
					loop_sts = 0;
					break;

			case	CMD_ERROR		:	// 顼׵
					Edit_RspFixData(&r_fbuf, STS_SUCCESS, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
					break;

			case	APM_CMD_GET_TIMER:	// Wake Up Timer μ
					if (r_fbuf.datalength != 0) {
						TraceDisp1("APM(%d) : apm_main() format error (%04X)", __LINE__, STS_FORMAT_ERROR);
						apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
						apm_error.OutputLog("APM(%d) : apm_main() format error (%04X)", __LINE__, STS_FORMAT_ERROR);
						Edit_RspFixData(&r_fbuf, STS_FORMAT_ERROR, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
						break;
					}

					w_vbuf = (unsigned char *)malloc(sizeof(struct apm_time));
					if (w_vbuf == NULL) {
						TraceDisp1("APM(%d) : apm_main() malloc error (%04X)", __LINE__, STS_MEMORY_ERROR);
						TraceDisp1("APM(%d) : apm_main() end", __LINE__);
						CmdErrorData_Edit(&r_fbuf, STS_MEMORY_ERROR);
						apm_Send_Data(fd.fd, (unsigned char *)&r_fbuf, FIXED_AREA_SIZE);
						pthread_exit((void *)1);
					}

					res = get_resume_timer(fd.apm_fd, (struct apm_time *)w_vbuf);
					if (res != SUCCESS) {
						TraceDisp1("APM(%d) : apm_main() error (%04X)", __LINE__, res);
						apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
						apm_error.OutputLog("APM(%d) : apm_main() error (%04X)", __LINE__, STS_UNKNOWN_COMMAND);
						Edit_RspFixData(&r_fbuf, STS_UNKNOWN_COMMAND, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
						break;
					}

					Edit_RspFixData(&r_fbuf, STS_SUCCESS, 
						0x0000, sizeof(struct apm_time), (FIXED_FORMAT *)w_buf, &w_len);

					w_len = reallocate(&w_buf, w_vbuf, sizeof(struct apm_time));

					break;

			case	APM_CMD_SET_TIMER:	// Wake Up Timer 
					if (r_fbuf.datalength != sizeof(struct apm_time)) {
						TraceDisp1("APM(%d) : apm_main() format error (%04X)", __LINE__, STS_FORMAT_ERROR);
						apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
						apm_error.OutputLog("APM(%d) : apm_main() format error (%04X)", __LINE__, STS_FORMAT_ERROR);
						Edit_RspFixData(&r_fbuf, STS_FORMAT_ERROR, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
						break;
					}

					res = set_resume_timer(fd.apm_fd, (struct apm_time *)r_vbuf);
					if (res != SUCCESS) {
						TraceDisp1("APM(%d) : apm_main() error (%04X)", __LINE__, res);
						apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
						apm_error.OutputLog("APM(%d) : apm_main() error (%04X)", __LINE__, STS_UNKNOWN_COMMAND);
						Edit_RspFixData(&r_fbuf, STS_UNKNOWN_COMMAND, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
						break;
					}

					memset(r_vbuf, 0, sizeof(struct apm_time));

					res = get_resume_timer(fd.apm_fd, (struct apm_time *)r_vbuf);
					if (res != SUCCESS) {
						TraceDisp1("APM(%d) : apm_main() error (%04X)", __LINE__, res);
						apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
						apm_error.OutputLog("APM(%d) : apm_main() error (%04X)", __LINE__, STS_UNKNOWN_COMMAND);
						break;
					}

					Edit_RspFixData(&r_fbuf, STS_SUCCESS, 
									0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);

					break;

/********** From Other Module *********/
			default:
					TraceDisp1("APM(%d) : apm_main() UNKNOWN_COMMAND error (%04X)", __LINE__, r_fbuf.command);
					apm_error.SetErrorCode(r_fbuf.command, __FILE__, __LINE__, errno);
					apm_error.OutputLog("APM(%d) : apm_main() UNKNOWN_COMMAND error (%04X)", __LINE__, r_fbuf.command);
					Edit_RspFixData(&r_fbuf, STS_UNKNOWN_COMMAND, 
								0x0000, 0, (FIXED_FORMAT *)w_buf, &w_len);
					break;
		}

		// ManagerؤΥǡ
		res = apm_Send_Data(fd.fd, w_buf, w_len);
		if (res != SUCCESS) {
			TraceDisp1("APM(%d) : apm_main() error (%04X)", __LINE__, res);
			TraceDisp1("APM(%d) : apm_main() end", __LINE__);
			apm_error.SetErrorCode(res, __FILE__, __LINE__, errno);
			apm_error.OutputLog("APM(%d) : apm_main() error (%04X)", __LINE__, res);
			pthread_exit((void *)1);
		}

		if (w_buf)  { free(w_buf);  w_buf  = NULL; }
		if (r_vbuf) { free(r_vbuf); r_vbuf = NULL; }

		TraceDisp2("APM(%d) : loop end", __LINE__);
	}

	// λ

	TraceDisp1("APM(%d) : apm_main() end", __LINE__);

	pthread_exit(0);
}


extern "C" {
// ؿ
pthread_t Initialize(int fd)
{
	pthread_t apm_thread;

	TraceDisp1("APM(%d) : fd = %04X", __LINE__, fd);
	pthread_create(&apm_thread, NULL, apm_main, (void *)fd);

	return apm_thread;
}

} /* end of extern "C" */
