/*
 * CopyrightC NEC Corporation, NEC CustomTechnica, Ltd. 2003
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "libapm_rw.h"
#include "myerror.h"
#include "debug.h"

#define SUCCESS		 0
#define TIMEOVER	-2

extern CErrorCode apm_error;

int apm_read(int fd, void *buf, int len)
{
	int		i = 0;		// read length
	int		r;			// rest
	unsigned char *ptr;	// buf pointer

	ptr = (unsigned char *)buf;

	while( i < len ) {
		r = read( fd, ptr+i, len-i);
		if( r < 0 )
			return r;
		if( r == 0 )
			return i;
		i+=r;
	}
	return i;
}

// ǡ 
int apm_Send_Data(int fd, unsigned char *obuf, unsigned long size)
{
	size_t		res;

	TraceDisp1("APM(%d) : apm_Send_Data() start", __LINE__);

	PrintFormat(obuf);

	res = write(fd, obuf, size);
	if( res != size ) {
		// 顼()
		TraceDisp1("APM(%d) : apm_Send_Data() write error (%d != %d)", __LINE__, size, res);
		TraceDisp1("APM(%d) : apm_Send_Data() end", __LINE__);
		apm_error.OutputLog("APM(%d) : apm_Send_Data() write error (%d != %d)", __LINE__, size, res);
		return FALSE;
	}

	TraceDisp1("APM(%d) : apm_Send_Data(len:%d) end", __LINE__, res);
	return SUCCESS;
}

// ǡɤ߹
int apm_Recv_vData(int fd, int r_vbuf_len, unsigned char *r_vbuf)
{
	int		res;	// ؿ

	TraceDisp1("APM(%d) : apm_Recv_vData() start", __LINE__);

	// ɤ߹
	res = apm_read(fd, r_vbuf, r_vbuf_len);
	if( res != r_vbuf_len ) {
		TraceDisp ("APM(%d) : apm_Recv_vData() read error (%d != %d)", __LINE__, r_vbuf_len, res);
		TraceDisp1("APM(%d) : apm_Recv_vData() end", __LINE__);
		apm_error.OutputLog("APM(%d) : apm_Recv_vData() read error (%d != %d)", __LINE__, r_vbuf_len, res);
		return FALSE;
	}

	TraceDisp1("APM(%d) : apm_Recv_vData() end", __LINE__);
	return SUCCESS;
}

// ǡɤ߹
int apm_Recv_fData(int fd, FIXED_FORMAT *r_fbuf)
{
	int		res;	// ؿ

	TraceDisp1("APM(%d) : apm_Recv_fData() start", __LINE__);

	// ɤ߹ΰν
	memset(r_fbuf, (char)NULL, FIXED_AREA_SIZE);

	// ɤ߹
	res = apm_read(fd, r_fbuf, FIXED_AREA_SIZE);
	if( res != FIXED_AREA_SIZE ) {
		// 顼()
		TraceDisp1("APM(%d) : apm_Recv_fData() read error (%d != %d)", __LINE__, FIXED_AREA_SIZE, res);
		TraceDisp1("APM(%d) : apm_Recv_fData() end", __LINE__);
		apm_error.OutputLog("APM(%d) : apm_Recv_fData() read error (%d != %d)", __LINE__, FIXED_AREA_SIZE, res);
		return FALSE;
	}

	TraceDisp1("APM(%d) : apm_Recv_fData() end", __LINE__);
	return SUCCESS;
}

// ǧ
int apm_Select_Data(int fd, struct timeval *timeout)
{       
	int		retsts;		// ؿ 
	fd_set	read_fds;	// read file descriptor set

	TraceDisp1("APM(%d) : apm_Select_Data() start", __LINE__);

	FD_ZERO(&read_fds);
	FD_SET(fd, &read_fds);

	retsts = select(fd+1, &read_fds, (fd_set *)0, (fd_set *)0, timeout);
	switch(retsts) {
		case	SELECT_TIMEOVER:	// ४С
					TraceDisp1("APM(%d) : apm_Select_Data() = SELECT_TIMEOVER", __LINE__);
					apm_error.OutputLog("APM(%d) : apm_Select_Data() = SELECT_TIMEOVER", __LINE__);
					break;
		case	SELECT_ERROR:		// 顼
					TraceDisp1("APM(%d) : apm_Select_Data() = SELECT_ERROR", __LINE__);
					apm_error.OutputLog("APM(%d) : apm_Select_Data() = SELECT_ERROR", __LINE__);
					break;
    }

	TraceDisp1("APM(%d) : apm_Select_Data() end", __LINE__);
    return retsts;
}


int apm_ReqSendRspRecv(int fd, void *i_buf, unsigned long i_len, void *o_fbuf, void **o_vbuf, unsigned long *o_vlen, struct timeval *timeout)
{
	int				res;
	FIXED_FORMAT	r_fbuf;
	unsigned char	*r_vbuf = NULL;

	TraceDisp1("APM(%d) : apm_ReqSendRspRecv() Start", __LINE__);

	res = apm_Send_Data(fd, (unsigned char *)i_buf, i_len);
	if( res != SUCCESS ) {
		TraceDisp1("APM(%d) : apm_ReqSendRspRecv() error (%04X)", __LINE__, STS_SELECT_ERROR);
		apm_error.OutputLog("APM(%d) : apm_ReqSendRspRecv() error (%04X)", __LINE__, res);
		return STS_SELECT_ERROR;
	}

	res = apm_Select_Data(fd, timeout);
	if( res != SUCCESS ) {
		TraceDisp1("APM(%d) : apm_ReqSendRspRecv() error (%04X)", __LINE__, res);
		apm_error.OutputLog("APM(%d) : apm_ReqSendRspRecv() error (%04X)", __LINE__, res);
		return res;
	}

	// ɤ߹ǡ¤ΤسǼ
	res = apm_Recv_fData(fd, &r_fbuf);
	if( res != SUCCESS ) {
		TraceDisp1("APM(%d) : apm_ReqSendRspRecv() error (%04X)", __LINE__, STS_READ_ERROR);
		apm_error.OutputLog("APM(%d) : apm_ReqSendRspRecv() error (%04X)", __LINE__, res);
		return STS_READ_ERROR;
	}

	memcpy(o_fbuf, &r_fbuf, FIXED_AREA_SIZE);

	if( r_fbuf.datalength > 0 ) {
		// Ĺ顢Υǡΰݤ
		r_vbuf = (unsigned char *)malloc(r_fbuf.datalength);
		if( r_vbuf == NULL ) {
			TraceDisp1("APM(%d) : apm_ReqSendRspRecv() malloc error", __LINE__);
			apm_error.OutputLog("APM(%d) : apm_ReqSendRspRecv() malloc error (%04X)", __LINE__, STS_MEMORY_ERROR);
			return STS_MEMORY_ERROR;
		}

		// ɤ߹ǡΰسǼ
		res = apm_Recv_vData(fd, r_fbuf.datalength, r_vbuf);
		if( res != SUCCESS ) {
			TraceDisp1("APM(%d) : apm_ReqSendRspRecv() error (%04X)", __LINE__, STS_READ_ERROR);
			apm_error.OutputLog("APM(%d) : apm_ReqSendRspRecv() error (%04X)", __LINE__, STS_READ_ERROR);
			free(r_vbuf);
			return STS_READ_ERROR;
		}

		*o_vlen = r_fbuf.datalength;
		(unsigned char *)*o_vbuf = r_vbuf;
	}


	TraceDisp1("APM(%d) : apm_ReqSendRspRecv(len:%d) End", __LINE__, *o_vlen);

	return SUCCESS;
}
