/*
 * CopyrightC NEC Corporation, NEC CustomTechnica, Ltd. 2003
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include "myerror.h"

const char*  szModuleName[] = { "Manager",
								"UI",
								"COM",
								"TV",
								"EPG",
								"FILE",
								"REC",
								"LCD",
								"SETUP",
								"RECOVERY",
								"UPDATE",
								"APM" };

#ifdef _DEBUG

void TraceDisp_main(const char* strtrace1)
{
	static char strtrace2[1024];
	struct tm* tmLocal;
	static time_t current_time;
	int nFileHandle;

	time( &current_time );
	tmLocal = localtime(&current_time);

	sprintf(strtrace2, "%02d/%02d/%02d %02d:%02d:%02d : %s\n",
						tmLocal->tm_year+1900,
						tmLocal->tm_mon+1,
						tmLocal->tm_mday,
						tmLocal->tm_hour,
						tmLocal->tm_min,
						tmLocal->tm_sec,
 						strtrace1);

	nFileHandle = open("logfile", O_APPEND | O_RDWR | O_CREAT , S_IRUSR | S_IWUSR);

	if(nFileHandle != 0)
	{
		write(nFileHandle, strtrace2, strlen(strtrace2));
		close(nFileHandle);
	}

	printf(strtrace2);

}

void TraceDisp(const char* szMsg, ...)
{
	static char strtrace1[1024];
	va_list argList;
	va_start(argList, szMsg);
	vsprintf(strtrace1, szMsg, argList);
	TraceDisp_main(strtrace1);
	va_end(argList);
}

#ifdef _DEBUG1
void TraceDisp1(const char* szMsg, ...)
{
	static char strtrace1[1024];
	va_list argList;
	va_start(argList, szMsg);
	vsprintf(strtrace1, szMsg, argList);
	TraceDisp_main(strtrace1);
	va_end(argList);
}
#endif

#ifdef _DEBUG2
void TraceDisp2(const char* szMsg, ...)
{
	static char strtrace1[1024];
	va_list argList;
	va_start(argList, szMsg);
	vsprintf(strtrace1, szMsg, argList);
	TraceDisp_main(strtrace1);
	va_end(argList);
}
#endif

#ifdef _DEBUG3
void TraceDisp3(const char* szMsg, ...)
{
	static char strtrace1[1024];
	va_list argList;
	va_start(argList, szMsg);
	vsprintf(strtrace1, szMsg, argList);
	TraceDisp_main(strtrace1);
	va_end(argList);
}
#endif

#endif

CSemaphore::CSemaphore()
{
	m_nSemId = 0;		// ޥեID
	m_keyid = 0;		// KEYID

}
CSemaphore::~CSemaphore()
{
	if(m_nSemId != 0)		// ޥեID
		DelSemValue();

	m_keyid = 0;		// KEYID
}

int CSemaphore::Initialize(int nModuleId)
{
	m_keyid 	= nModuleId;
	m_nModuleId = nModuleId;
	if((m_nSemId = semget((key_t)m_keyid, 1, 666|IPC_CREAT)) == -1)
		return -1;

	return SetSemValue();
}


int CSemaphore::SetSemValue()
{
	union semun sem_union;

	sem_union.val = 1;
	if(semctl(m_nSemId, 0, SETVAL, sem_union) == -1)
		return -1;

	return 0;
}

int CSemaphore::DelSemValue()
{
	union semun sem_union;

	sem_union.val = 1;
	if(semctl(m_nSemId, 0, IPC_RMID, sem_union) == -1)
		return -1;

	return 0;
}

int CSemaphore::Semaphore_p()
{
	struct sembuf sem_b;
	sem_b.sem_num = 0;
	sem_b.sem_op  = -1;
	sem_b.sem_flg = SEM_UNDO;

	if(semop(m_nSemId, &sem_b, 1) == -1)
		return -1;

	return 0;
}

int CSemaphore::Semaphore_v()
{
	struct sembuf sem_b;
	sem_b.sem_num = 0;
	sem_b.sem_op  = 1;
	sem_b.sem_flg = SEM_UNDO;

	if(semop(m_nSemId, &sem_b, 1) == -1)
		return -1;

	return 0;
}


int CSemaphore::Lock()
{
	return Semaphore_p();
}


int CSemaphore::Unlock()
{
	return Semaphore_v();
}



CErrorCode::CErrorCode()
{
	m_nErrorCode = NO_ERROR;				// 顼
	memset(m_szErrorFile, 0, MAX_PATH);	// 顼ȯե̾
	m_nLineNo= 0;							// 顼ȯֹ
	m_nLastError  = 0;						// 顼ȯֹ
	memset(m_szErrorLogFile, 0, MAX_PATH);	// 顼ե̾
	m_nErrorLogLine    = 0;					// 顼񤭹Կ
	m_nErrorLogLineMax = 0;					// 顼񤭹Կ
	m_nErrorLogColumsMax = 0;				// 顼񤭹
	// ƥ륻
//	InitializeCriticalSection(&m_csAcressFileDB);
}

CErrorCode::~CErrorCode()
{
	// ƥ륻Ѵ
//	DeleteCriticalSection(&m_csAcressFileDB);

}

//
//	顼ν
//
//
//	const char* szLogName ... ե̾
//	int nLine ............... Կ (ޤࡢǥå)
//	int nModuleId ........... ⥸塼IDֹ
//
int CErrorCode::Initialize(const char* szLogName, int nLine, int nColums, int nModuleId)
{
	char  buffer[MAX_PATH];
	char  work[MAX_PATH];
	int   hFile = 0;
	int   nRet = 0;

	// ե̾
	strcpy(m_szErrorLogFile, szLogName);

	// Կ
	m_nErrorLogLineMax = nLine;

	// 
	m_nErrorLogColumsMax = nColums;
	if(m_nErrorLogColumsMax > MAX_PATH)
		m_nErrorLogColumsMax = MAX_PATH;

	// ⥸塼ID
	m_nModuleId = nModuleId;

	m_Semaphore.Initialize(m_nModuleId*0x1000);

	memset(buffer, ' ', MAX_PATH-1);

	m_Semaphore.Lock();
	while(TRUE)
	{
		int nWriteSize;
		struct stat status;
		BOOL fInit = TRUE;

		stat(m_szErrorLogFile, &status);

		// ȹֹxMAX_PATHפ롩
		if(status.st_size == m_nErrorLogLineMax*m_nErrorLogColumsMax)
		{
			int nReadSize;
			hFile = open(m_szErrorLogFile, O_APPEND | O_RDWR | O_CREAT,  S_IRUSR | S_IWUSR, 0777);
			if(hFile == -1)
			{
				nRet = -1;
				break;
			}
			// ܤɤ߹ߡ
			if(lseek(hFile, 0, SEEK_SET) == -1)
			{
				nRet = -1;
				break;
			}

			// ܤɤ
			if((nReadSize = read(hFile, buffer, m_nErrorLogColumsMax)) == -1)
			{
				nRet = -1;
				break;
			}

			if(nReadSize == m_nErrorLogColumsMax)
			{
				// 1ܤϸߤιԿȺԿäƤ
				int nErrorLogLineMax = 0;
				int nErrorLogColumsMax = 0;
				sscanf(buffer, "%d,%d,%d", &m_nErrorLogLine, &nErrorLogLineMax, &nErrorLogColumsMax);

				if((m_nErrorLogLineMax == nErrorLogLineMax) &&
					(m_nErrorLogLineMax > m_nErrorLogLine))
					fInit = FALSE;

				if(m_nErrorLogColumsMax == nErrorLogColumsMax)
					fInit = FALSE;
			}
		}
		if(hFile > 0)
		{
			close(hFile);
			hFile = 0;
		}

		// ɬפäȤե뤬ʤȤ䡢۾äȤʤɡ
		if(fInit)
		{
			//	顼ȥ졼ե򿷵˺
			hFile = open(m_szErrorLogFile, O_TRUNC | O_WRONLY | O_CREAT , 0777);
			if(hFile == -1)
			{
				nRet = -1;
				break;
			}
			m_nErrorLogLine = 1;
			sprintf(work, "%d,%d,%d", m_nErrorLogLine, m_nErrorLogLineMax, m_nErrorLogColumsMax);
			memcpy(buffer, work, strlen(work));
			buffer[m_nErrorLogColumsMax-2] = '\x0d';
			buffer[m_nErrorLogColumsMax-1] = '\x0a';

			if(lseek(hFile, 0, SEEK_SET) == -1)
			{
				nRet = -1;
				break;
			}
			if((nWriteSize = write(hFile, buffer, m_nErrorLogColumsMax)) == -1)
			{
				nRet = -1;
				break;
			}
			if(nWriteSize != m_nErrorLogColumsMax)
			{
				nRet = -1;
				break;
			}
			memset(buffer, ' ', MAX_PATH);
			buffer[m_nErrorLogColumsMax-2] = '\x0d';
			buffer[m_nErrorLogColumsMax-1] = '\x0a';

			// ڡιԤԿʬ񤭽Ф
			for(int i=0; i<m_nErrorLogLineMax-1; i++)
			{
				if((nWriteSize = write(hFile, buffer, m_nErrorLogColumsMax)) == -1)
				{
					nRet = -1;
					break;
				}
				if(nWriteSize != m_nErrorLogColumsMax)
				{
					nRet = -1;
					break;
				}
			}
		}
		break;
	}
	if(hFile > 0)
		close(hFile);

	m_Semaphore.Unlock();

	return nRet;
}


//
//	顼å
//
int CErrorCode::SetErrorCode(int errcode, const char* szErrorFile, int nLineNo, int nLastError)
{
	int nRet = 0;
	// ƥ륻
//	EnterCriticalSection(&m_csAcressFileDB);
	m_Semaphore.Lock();

	m_nErrorCode  = errcode;
	m_nLastError = nLastError;
	m_nLineNo	  = nLineNo;
	strcpy(m_szErrorFile, szErrorFile);

	TraceDisp("Error code=0x%x File=%s Line=%d", errcode, szErrorFile, nLineNo);
	nRet  = OutputValist("Error code=0x%x File=%s Line=%d", errcode, szErrorFile, nLineNo);

//	LeaveCriticalSection(&m_csAcressFileDB);
	m_Semaphore.Unlock();
	return nRet;
}

int CErrorCode::OutputLog(char* szMsg, ...)
{
	va_list argList;
	int nret;
	m_Semaphore.Lock();
	va_start(argList, szMsg);
	nret = ErrorLogOut(szMsg, argList);
	va_end(argList);
	m_Semaphore.Unlock();
	return	nret;
}

int CErrorCode::OutputValist(char* szMsg, ...)
{
	va_list argList;
	int nret;
	va_start(argList, szMsg);
	nret = ErrorLogOut(szMsg, argList);
	va_end(argList);
	return	nret;
}

//
//	顼
//
int CErrorCode::ErrorLogOut(char* szMsg, va_list argList)
{
	static char strtrace1[2048];
	static char strtrace2[2048];
	char buffer[MAX_PATH];
	char work[MAX_PATH];
	static time_t current_time;
	struct tm* tmLocal;
	int nWriteSize;
	int nRet = NO_ERROR;
	int hFile = 0;

	time( &current_time );
	tmLocal = localtime(&current_time);

	vsprintf(strtrace1, szMsg, argList);
	sprintf(strtrace2, "%02d/%02d/%02d %02d:%02d:%02d %s: %s",
						tmLocal->tm_year+1900,
						tmLocal->tm_mon+1,
						tmLocal->tm_mday,
						tmLocal->tm_hour,
						tmLocal->tm_min,
						tmLocal->tm_sec,
						szModuleName[m_nModuleId],
						strtrace1);

	memset(buffer, ' ', MAX_PATH-1);

	while(TRUE)
	{
		if((hFile = open(m_szErrorLogFile, O_RDWR | O_CREAT , S_IRUSR | S_IWUSR)) == -1)
		{
			nRet = -1;
			break;
		}

		// 顼
		memcpy(buffer, strtrace2, strlen(strtrace2));
		buffer[m_nErrorLogColumsMax-2] = '\x0d';
		buffer[m_nErrorLogColumsMax-1] = '\x0a';

		if(lseek(hFile, m_nErrorLogLine*m_nErrorLogColumsMax, SEEK_SET) == -1)
		{
			nRet = -1;
			break;
		}

		if((nWriteSize = write(hFile, buffer, m_nErrorLogColumsMax)) == -1)
		{
			nRet = -1;
			break;
		}

		// ܤ򹹿  ˽񤭹,
		m_nErrorLogLine++;

		if(m_nErrorLogLine == m_nErrorLogLineMax)
			m_nErrorLogLine = 1;
		sprintf(work, "%d,%d,%d", m_nErrorLogLine, m_nErrorLogLineMax, m_nErrorLogColumsMax);
		memset(buffer, ' ', MAX_PATH-1);
		memcpy(buffer, work, strlen(work));
		buffer[m_nErrorLogColumsMax-2] = '\x0d';
		buffer[m_nErrorLogColumsMax-1] = '\x0a';
		if(lseek(hFile, 0, SEEK_SET) == -1)
		{
			nRet = -1;
			break;
		}
		if((nWriteSize = write(hFile, buffer, m_nErrorLogColumsMax)) == -1)
		{
			nRet = -1;
			break;
		}

		break;
	}
	if(hFile > 0)
		close(hFile);

	return nRet;
}


#ifdef TEST

CErrorCode myerror;	// ͣΥ顼֥

int main()
{
	myerror.Initialize("./errorlog.txt", 128, 80, 1);
	myerror.SetErrorCode(0x1234, __FILE__, __LINE__, errno);
	myerror.SetErrorCode(0x2345, __FILE__, __LINE__, errno);
	myerror.OutputLog("error = %d, FILE=%s LINE=%d", myerror.GetErrorCode(), __FILE__, __LINE__);
	return 0;
}

#endif
